@@ -1,5 +1,5 @@
 use crate::net::config::NetworkConfig;
-use crate::net::peers::{AddressBook, AddressSource, Event, PeerInfo};
+use crate::net::peers::{AddressBook, AddressSource, Event, PeerInfo, SwarmEvents};
 use fnv::FnvHashMap;
 use futures::channel::{mpsc, oneshot};
 use futures::stream::Stream;
@@ -23,10 +23,12 @@ use libp2p::swarm::NetworkBehaviourEventProcess;
 use libp2p::NetworkBehaviour;
 use libp2p::{Multiaddr, PeerId};
 use libp2p_bitswap::{Bitswap, BitswapEvent, BitswapStore};
+use libp2p_blake_streams::{StreamSync, StreamSyncEvent};
 use libp2p_broadcast::{Broadcast, BroadcastEvent, Topic};
-use libp2p_quic::ToLibp2p;
+use libp2p_quic::{PublicKey, ToLibp2p};
 use prometheus::Registry;
 use std::collections::HashSet;
+use std::sync::Arc;
 use thiserror::Error;
 
 #[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
@@ -79,7 +81,12 @@ enum QueryChannel {
     GetRecord(oneshot::Sender<Result<Vec<PeerRecord>>>),
     PutRecord(oneshot::Sender<Result<()>>),
 }
-
+#[derive(Clone, Debug, PartialEq)]
+pub enum GossipEvent {
+    Subscribed(PeerId),
+    Message(PeerId, Arc<[u8]>),
+    Unsubscribed(PeerId),
+}
 /// Behaviour type.
 #[derive(NetworkBehaviour)]
 pub struct NetworkBackendBehaviour<P: StoreParams> {
@@ -91,13 +98,14 @@ pub struct NetworkBackendBehaviour<P: StoreParams> {
     bitswap: Toggle<Bitswap<P>>,
     gossipsub: Toggle<Gossipsub>,
     broadcast: Toggle<Broadcast>,
+    streams: Toggle<StreamSync>,
 
     #[behaviour(ignore)]
     bootstrap_complete: bool,
     #[behaviour(ignore)]
     queries: FnvHashMap<QueryId, QueryChannel>,
     #[behaviour(ignore)]
-    subscriptions: FnvHashMap<String, Vec<mpsc::UnboundedSender<Vec<u8>>>>,
+    subscriptions: FnvHashMap<String, Vec<mpsc::UnboundedSender<GossipEvent>>>,
 }
 
 impl<P: StoreParams> NetworkBehaviourEventProcess<MdnsEvent> for NetworkBackendBehaviour<P> {
@@ -154,7 +162,7 @@ pub struct KadPutRecordError(pub libp2p::kad::PutRecordError);
 impl<P: StoreParams> NetworkBehaviourEventProcess<KademliaEvent> for NetworkBackendBehaviour<P> {
     fn inject_event(&mut self, event: KademliaEvent) {
         tracing::trace!("kademlia event {:?}", event);
-        if let KademliaEvent::QueryResult { id, result, .. } = event {
+        if let KademliaEvent::OutboundQueryCompleted { id, result, .. } = event {
             match result {
                 QueryResult::Bootstrap(Ok(BootstrapOk { num_remaining, .. })) => {
                     tracing::trace!("remaining {}", num_remaining);
@@ -163,6 +171,7 @@ impl<P: StoreParams> NetworkBehaviourEventProcess<KademliaEvent> for NetworkBack
                         if let Some(QueryChannel::Bootstrap(ch)) = self.queries.remove(&id.into()) {
                             ch.send(Ok(())).ok();
                         }
+                        self.peers.notify(Event::Bootstrapped);
                     }
                 }
                 QueryResult::Bootstrap(Err(err)) => {
@@ -272,14 +281,14 @@ impl<P: StoreParams> NetworkBehaviourEventProcess<PingEvent> for NetworkBackendB
                 peer,
                 result: Result::Ok(PingSuccess::Ping { rtt }),
             } => {
-                tracing::trace!("ping: rtt to {} is {} ms", peer, rtt.as_millis());
+                //tracing::trace!("ping: rtt to {} is {} ms", peer, rtt.as_millis());
                 self.peers.set_rtt(&peer, Some(rtt));
             }
             PingEvent {
-                peer,
+                peer: _,
                 result: Result::Ok(PingSuccess::Pong),
             } => {
-                tracing::trace!("ping: pong from {}", peer);
+                //tracing::trace!("ping: pong from {}", peer);
             }
             PingEvent {
                 peer,
@@ -325,25 +334,31 @@ impl<P: StoreParams> NetworkBehaviourEventProcess<GossipsubEvent> for NetworkBac
     fn inject_event(&mut self, event: GossipsubEvent) {
         match event {
             GossipsubEvent::Message {
-                message: GossipsubMessage { data, topic, .. },
+                message:
+                    GossipsubMessage {
+                        data,
+                        topic,
+                        source,
+                        ..
+                    },
+                propagation_source,
                 ..
             } => {
-                if let Some(subscribers) = self.subscriptions.get_mut(topic.as_str()) {
-                    subscribers
-                        .retain(|subscriber| subscriber.unbounded_send(data.clone()).is_ok());
-                    if subscribers.is_empty() {
-                        self.unsubscribe(topic.as_str());
-                        self.subscriptions.remove(topic.as_str());
-                    }
-                }
+                self.notify_subscribers(
+                    &*topic.to_string(),
+                    GossipEvent::Message(source.unwrap_or(propagation_source), data.into()),
+                );
             }
             GossipsubEvent::Subscribed { peer_id, topic, .. } => {
                 self.peers
                     .notify(Event::Subscribed(peer_id, topic.to_string()));
+
+                self.notify_subscribers(&*topic.to_string(), GossipEvent::Subscribed(peer_id));
             }
             GossipsubEvent::Unsubscribed { peer_id, topic, .. } => {
                 self.peers
                     .notify(Event::Unsubscribed(peer_id, topic.to_string()));
+                self.notify_subscribers(&*topic.to_string(), GossipEvent::Unsubscribed(peer_id));
             }
         }
     }
@@ -352,32 +367,37 @@ impl<P: StoreParams> NetworkBehaviourEventProcess<GossipsubEvent> for NetworkBac
 impl<P: StoreParams> NetworkBehaviourEventProcess<BroadcastEvent> for NetworkBackendBehaviour<P> {
     fn inject_event(&mut self, event: BroadcastEvent) {
         match event {
-            BroadcastEvent::Received(_peer_id, topic, data) => {
+            BroadcastEvent::Received(peer_id, topic, data) => {
                 let topic = std::str::from_utf8(&topic).unwrap();
-                if let Some(subscribers) = self.subscriptions.get_mut(topic) {
-                    subscribers
-                        .retain(|subscriber| subscriber.unbounded_send(data.to_vec()).is_ok());
-                    if subscribers.is_empty() {
-                        self.unsubscribe(topic);
-                        self.subscriptions.remove(topic);
-                    }
-                }
+                self.notify_subscribers(topic, GossipEvent::Message(peer_id, data));
             }
             BroadcastEvent::Subscribed(peer_id, topic) => {
                 if let Ok(topic) = std::str::from_utf8(&topic) {
                     self.peers.notify(Event::Subscribed(peer_id, topic.into()));
+                    self.notify_subscribers(topic, GossipEvent::Subscribed(peer_id));
                 }
             }
             BroadcastEvent::Unsubscribed(peer_id, topic) => {
                 if let Ok(topic) = std::str::from_utf8(&topic) {
                     self.peers
                         .notify(Event::Unsubscribed(peer_id, topic.into()));
+                    self.notify_subscribers(topic, GossipEvent::Unsubscribed(peer_id));
                 }
             }
         }
     }
 }
 
+impl<P: StoreParams> NetworkBehaviourEventProcess<StreamSyncEvent> for NetworkBackendBehaviour<P> {
+    fn inject_event(&mut self, event: StreamSyncEvent) {
+        match event {
+            StreamSyncEvent::NewHead(head) => {
+                self.peers.notify(Event::NewHead(head));
+            }
+        }
+    }
+}
+
 impl<P: StoreParams> NetworkBehaviourEventProcess<void::Void> for NetworkBackendBehaviour<P> {
     fn inject_event(&mut self, _event: void::Void) {}
 }
@@ -388,6 +408,7 @@ impl<P: StoreParams> NetworkBackendBehaviour<P> {
         config: &mut NetworkConfig,
         store: S,
     ) -> Result<Self> {
+        let public = config.node_key.public;
         let node_key = config.node_key.to_keypair();
         let node_name = config.node_name.clone();
         let peer_id = node_key.public().into_peer_id();
@@ -422,9 +443,14 @@ impl<P: StoreParams> NetworkBackendBehaviour<P> {
             .bitswap
             .take()
             .map(|config| Bitswap::new(config, store));
+        let streams = if let Some(config) = config.streams.take() {
+            Some(StreamSync::new(config)?)
+        } else {
+            None
+        };
         Ok(Self {
             bootstrap_complete: false,
-            peers: AddressBook::new(peer_id, node_name),
+            peers: AddressBook::new(peer_id, node_name, public),
             mdns: mdns.into(),
             kad: kad.into(),
             ping: ping.into(),
@@ -432,11 +458,16 @@ impl<P: StoreParams> NetworkBackendBehaviour<P> {
             bitswap: bitswap.into(),
             gossipsub: gossipsub.into(),
             broadcast: broadcast.into(),
+            streams: streams.into(),
             queries: Default::default(),
             subscriptions: Default::default(),
         })
     }
 
+    pub fn local_public_key(&self) -> &PublicKey {
+        self.peers.local_public_key()
+    }
+
     pub fn local_node_name(&self) -> &str {
         self.peers.local_node_name()
     }
@@ -492,6 +523,10 @@ impl<P: StoreParams> NetworkBackendBehaviour<P> {
         rx
     }
 
+    pub fn is_bootstrapped(&self) -> bool {
+        self.bootstrap_complete
+    }
+
     pub fn get_closest_peers<K>(&mut self, key: K) -> GetClosestPeersChannel
     where
         K: Into<BucketKey<K>> + Into<Vec<u8>> + Clone,
@@ -587,7 +622,7 @@ impl<P: StoreParams> NetworkBackendBehaviour<P> {
         }
     }
 
-    pub fn subscribe(&mut self, topic: &str) -> Result<impl Stream<Item = Vec<u8>>> {
+    pub fn subscribe(&mut self, topic: &str) -> Result<impl Stream<Item = GossipEvent>> {
         if self.gossipsub.as_ref().is_none() && self.broadcast.as_ref().is_none() {
             return Err(DisabledProtocol.into());
         }
@@ -624,6 +659,16 @@ impl<P: StoreParams> NetworkBackendBehaviour<P> {
         }
     }
 
+    fn notify_subscribers(&mut self, topic: &str, event: GossipEvent) {
+        if let Some(subscribers) = self.subscriptions.get_mut(topic) {
+            subscribers.retain(|subscriber| subscriber.unbounded_send(event.clone()).is_ok());
+            if subscribers.is_empty() {
+                self.unsubscribe(topic);
+                self.subscriptions.remove(topic);
+            }
+        }
+    }
+
     pub fn publish(&mut self, topic: &str, msg: Vec<u8>) -> Result<()> {
         use libp2p::gossipsub::error::PublishError;
         if let Some(gossipsub) = self.gossipsub.as_mut() {
@@ -692,7 +737,11 @@ impl<P: StoreParams> NetworkBackendBehaviour<P> {
         Ok(())
     }
 
-    pub fn event_stream(&mut self) -> impl Stream<Item = Event> {
-        self.peers.event_stream()
+    pub fn swarm_events(&mut self) -> SwarmEvents {
+        self.peers.swarm_events()
+    }
+
+    pub fn streams(&mut self) -> &mut StreamSync {
+        self.streams.as_mut().expect("streams enabled")
     }
 }
