@@ -9,9 +9,12 @@ use libp2p::multiaddr::Protocol;
 use libp2p::swarm::protocols_handler::DummyProtocolsHandler;
 use libp2p::swarm::{DialPeerCondition, NetworkBehaviour, NetworkBehaviourAction, PollParameters};
 use libp2p::{Multiaddr, PeerId};
+use libp2p_blake_streams::Head;
+use libp2p_quic::PublicKey;
 use prometheus::{IntCounter, IntGauge, Registry};
 use std::borrow::Cow;
 use std::collections::VecDeque;
+use std::pin::Pin;
 use std::task::{Context, Poll};
 use std::time::Duration;
 
@@ -29,6 +32,8 @@ pub enum Event {
     Disconnected(PeerId),
     Subscribed(PeerId, String),
     Unsubscribed(PeerId, String),
+    NewHead(Head),
+    Bootstrapped,
 }
 
 #[derive(Clone, Debug, Default, Eq, PartialEq)]
@@ -116,10 +121,26 @@ fn normalize_addr_ref<'a>(addr: &'a Multiaddr, peer: &PeerId) -> Cow<'a, Multiad
     }
 }
 
+trait MultiaddrExt {
+    fn is_loopback(&self) -> bool;
+}
+
+impl MultiaddrExt for Multiaddr {
+    fn is_loopback(&self) -> bool {
+        if let Some(Protocol::Ip4(addr)) = self.iter().next() {
+            if !addr.is_loopback() {
+                return false;
+            }
+        }
+        true
+    }
+}
+
 #[derive(Debug)]
 pub struct AddressBook {
     local_node_name: String,
     local_peer_id: PeerId,
+    local_public_key: PublicKey,
     peers: FnvHashMap<PeerId, PeerInfo>,
     connections: FnvHashMap<PeerId, Multiaddr>,
     event_stream: Vec<mpsc::UnboundedSender<Event>>,
@@ -127,10 +148,15 @@ pub struct AddressBook {
 }
 
 impl AddressBook {
-    pub fn new(local_peer_id: PeerId, local_node_name: String) -> Self {
+    pub fn new(
+        local_peer_id: PeerId,
+        local_node_name: String,
+        local_public_key: PublicKey,
+    ) -> Self {
         Self {
             local_node_name,
             local_peer_id,
+            local_public_key,
             peers: Default::default(),
             connections: Default::default(),
             event_stream: Default::default(),
@@ -138,6 +164,10 @@ impl AddressBook {
         }
     }
 
+    pub fn local_public_key(&self) -> &PublicKey {
+        &self.local_public_key
+    }
+
     pub fn local_node_name(&self) -> &str {
         &self.local_node_name
     }
@@ -162,6 +192,9 @@ impl AddressBook {
         if peer == self.local_peer_id() {
             return;
         }
+        if address.is_loopback() {
+            return;
+        }
         tracing::trace!(
             "adding address {} for peer {} from {:?}",
             address,
@@ -215,10 +248,10 @@ impl AddressBook {
         }
     }
 
-    pub fn event_stream(&mut self) -> impl Stream<Item = Event> {
+    pub fn swarm_events(&mut self) -> SwarmEvents {
         let (tx, rx) = mpsc::unbounded();
         self.event_stream.push(tx);
-        rx
+        SwarmEvents(rx)
     }
 
     pub fn notify(&mut self, event: Event) {
@@ -241,6 +274,16 @@ impl AddressBook {
     }
 }
 
+pub struct SwarmEvents(mpsc::UnboundedReceiver<Event>);
+
+impl Stream for SwarmEvents {
+    type Item = Event;
+
+    fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context) -> Poll<Option<Self::Item>> {
+        Pin::new(&mut self.0).poll_next(cx)
+    }
+}
+
 impl NetworkBehaviour for AddressBook {
     type ProtocolsHandler = DummyProtocolsHandler;
     type OutEvent = void::Void;
@@ -396,12 +439,13 @@ impl NetworkBehaviour for AddressBook {
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::generate_keypair;
     use futures::stream::StreamExt;
 
     #[async_std::test]
     async fn test_dial_basic() {
-        let mut book = AddressBook::new(PeerId::random(), "".into());
-        let mut stream = book.event_stream();
+        let mut book = AddressBook::new(PeerId::random(), "".into(), generate_keypair().public);
+        let mut stream = book.swarm_events();
         let peer_a = PeerId::random();
         let addr_1: Multiaddr = "/ip4/1.1.1.1/tcp/3333".parse().unwrap();
         let mut addr_1_2 = addr_1.clone();
@@ -424,8 +468,8 @@ mod tests {
 
     #[async_std::test]
     async fn test_dial_with_added_addrs() {
-        let mut book = AddressBook::new(PeerId::random(), "".into());
-        let mut stream = book.event_stream();
+        let mut book = AddressBook::new(PeerId::random(), "".into(), generate_keypair().public);
+        let mut stream = book.swarm_events();
         let peer_a = PeerId::random();
         let addr_1: Multiaddr = "/ip4/1.1.1.1/tcp/3333".parse().unwrap();
         let addr_2: Multiaddr = "/ip4/2.2.2.2/tcp/3333".parse().unwrap();
